# Copyright (C) 2024. All rights reserved.
"""v3.0.0 specific class definitions"""

import ctypes
import logging

from dlt.core.core_base import (
    dltlib,
    cDltStorageHeader,
    cDltStandardHeader,
    cDltStandardHeaderExtra,
    cDltExtendedHeader,
)

# noqa: F401
from dlt.core.core_21810 import (
    DLT_CLIENT_MODE_UNDEFINED,
    DLT_CLIENT_MODE_TCP,
    DLT_CLIENT_MODE_SERIAL,
    DLT_CLIENT_MODE_UNIX,
    DLT_CLIENT_MODE_UDP_MULTICAST,
    DLT_RECEIVE_SOCKET,
    DLT_RECEIVE_UDP_SOCKET,
    DLT_RECEIVE_FD,
    DLT_ID_SIZE,
    DLT_FILTER_MAX,
    DLT_RETURN_ERROR,
    MAX_FILTER_REACHED,
    REPEATED_FILTER,
    sockaddr_in,
    cDltReceiver,
)

__all__ = [
    "cDltClient",
    "cDLTFilter",
    "cDLTMessage",
    "DLT_CLIENT_MODE_UNDEFINED",
    "DLT_CLIENT_MODE_TCP",
    "DLT_CLIENT_MODE_SERIAL",
    "DLT_CLIENT_MODE_UNIX",
    "DLT_CLIENT_MODE_UDP_MULTICAST",
    "DLT_RECEIVE_SOCKET",
    "DLT_RECEIVE_UDP_SOCKET",
    "DLT_RECEIVE_FD",
    "DLT_ID_SIZE",
    "DLT_FILTER_MAX",
    "DLT_RETURN_ERROR",
    "MAX_FILTER_REACHED",
    "REPEATED_FILTER",
    "sockaddr_in",
    "cDltReceiver",
]

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class cDLTMessage(ctypes.Structure):
    """The structure of the DLT messages. Packed in v3."""

    _fields_ = [
        ("found_serialheader", ctypes.c_int8),
        ("resync_offset", ctypes.c_int32),
        ("headersize", ctypes.c_int32),
        ("datasize", ctypes.c_int32),
        (
            "headerbuffer",
            ctypes.c_uint8
            * (
                ctypes.sizeof(cDltStorageHeader)
                + ctypes.sizeof(cDltStandardHeader)
                + ctypes.sizeof(cDltStandardHeaderExtra)
                + ctypes.sizeof(cDltExtendedHeader)
            ),
        ),
        ("databuffer", ctypes.POINTER(ctypes.c_uint8)),
        ("databuffersize", ctypes.c_uint32),
        ("p_storageheader", ctypes.POINTER(cDltStorageHeader)),
        ("p_standardheader", ctypes.POINTER(cDltStandardHeader)),
        ("headerextra", cDltStandardHeaderExtra),
        ("p_extendedheader", ctypes.POINTER(cDltExtendedHeader)),
    ]
    _pack_ = 1
    _layout_ = "ms"


class cDltClient(ctypes.Structure):  # pylint: disable=invalid-name
    """
    typedef struct
    {
        DltReceiver receiver;      /**< receiver pointer to dlt receiver structure */
        int sock;                  /**< sock Connection handle/socket */
        char *servIP;              /**< servIP IP adress/Hostname of interface */
        char *hostip;              /**< hostip IP address of UDP host receiver interface */
        uint16_t  port;            /**< Port for TCP connections (optional) */
        char *serialDevice;        /**< serialDevice Devicename of serial device */
        char *socketPath;          /**< socketPath Unix socket path */
        char ecuid[4];             /**< ECU id */
        uint8_t ecuid2len;         /**< Version 2 ECU id length */
        char *ecuid2;              /**< Version 2 ECU id of variable length*/
        speed_t baudrate;          /**< baudrate Baudrate of serial interface, as speed_t */
        int mode;                  /**< mode DltClientMode */
        int send_serial_header;    /**< (Boolean) Send DLT messages with serial header */
        int resync_serial_header;  /**< (Boolean) Resync to serial header on all connection */
    } DltClient;
    """

    _fields_ = [
        ("receiver", cDltReceiver),
        ("sock", ctypes.c_int),
        ("servIP", ctypes.c_char_p),
        ("hostip", ctypes.c_char_p),
        ("port", ctypes.c_uint16),
        ("serialDevice", ctypes.c_char_p),
        ("socketPath", ctypes.c_char_p),
        ("ecuid", ctypes.c_char * 4),
        ("ecuid2len", ctypes.c_uint8),
        ("ecuid2", ctypes.c_char_p),
        ("baudrate", ctypes.c_uint),
        ("mode", ctypes.c_int),
        ("send_serial_header", ctypes.c_int),
        ("resync_serial_header", ctypes.c_int),
    ]


class cDLTFilter(ctypes.Structure):  # pylint: disable=invalid-name
    """
    typedef struct
    {
        char apid[DLT_FILTER_MAX][DLT_ID_SIZE]; /**< application id */
        char ctid[DLT_FILTER_MAX][DLT_ID_SIZE]; /**< context id */
        uint8_t apid2len[DLT_FILTER_MAX];       /**< length of application id */
        char *apid2[DLT_FILTER_MAX];            /**< application id */
        uint8_t ctid2len[DLT_FILTER_MAX];       /**< length of context id */
        char *ctid2[DLT_FILTER_MAX];            /**< context id */
        int log_level[DLT_FILTER_MAX];          /**< log level */
        int32_t payload_max[DLT_FILTER_MAX];    /**< upper border for payload */
        int32_t payload_min[DLT_FILTER_MAX];    /**< lower border for payload */
        int counter;                            /**< number of filters */
    } DltFilter;
    """

    _fields_ = [
        ("apid", (ctypes.c_char * DLT_ID_SIZE) * DLT_FILTER_MAX),
        ("ctid", (ctypes.c_char * DLT_ID_SIZE) * DLT_FILTER_MAX),
        ("apid2len", ctypes.c_uint8 * DLT_FILTER_MAX),
        ("apid2", ctypes.c_char_p * DLT_FILTER_MAX),
        ("ctid2len", ctypes.c_uint8 * DLT_FILTER_MAX),
        ("ctid2", ctypes.c_char_p * DLT_FILTER_MAX),
        ("log_level", ctypes.c_int * DLT_FILTER_MAX),
        ("payload_max", (ctypes.c_int32 * DLT_FILTER_MAX)),
        ("payload_min", (ctypes.c_int32 * DLT_FILTER_MAX)),
        ("counter", ctypes.c_int),
    ]

    # pylint: disable=too-many-arguments
    def add(self, apid, ctid, log_level=0, payload_min=0, payload_max=ctypes.c_uint32(-1).value // 2):
        """Add new filter pair"""
        if isinstance(apid, str):
            apid = bytes(apid, "ascii")
        if isinstance(ctid, str):
            ctid = bytes(ctid, "ascii")
        if (
            dltlib.dlt_filter_add(
                ctypes.byref(self), apid or b"", ctid or b"", log_level, payload_min, payload_max, self.verbose
            )
            == DLT_RETURN_ERROR
        ):
            if self.counter >= DLT_FILTER_MAX:
                logger.error("Maximum number (%d) of allowed filters reached, ignoring filter!\n", DLT_FILTER_MAX)
                return MAX_FILTER_REACHED
            logger.debug("Filter ('%s', '%s') already exists", apid, ctid)
            return REPEATED_FILTER
        return 0



# Copyright (C) 2015. BMW Car IT GmbH. All rights reserved.
"""Basic unittests for DLTFilter definition"""

import unittest

import ctypes

from dlt.dlt import DLTFilter
from dlt.core.core_base import DLT_FILTER_MAX, DLT_ID_SIZE


class TestDLTFilter(unittest.TestCase):
    def setUp(self):
        self.dlt_filter = DLTFilter()

    def tearDown(self):
        del self.dlt_filter

    def test_init(self):
        assert len(self.dlt_filter.apid) == DLT_FILTER_MAX
        assert len(self.dlt_filter.ctid) == DLT_FILTER_MAX
        assert self.dlt_filter.counter == 0

        for entry in self.dlt_filter.apid:
            assert ctypes.string_at(entry, DLT_ID_SIZE) == b"\0\0\0\0"

        for entry in self.dlt_filter.ctid:
            assert ctypes.string_at(entry, DLT_ID_SIZE) == b"\0\0\0\0"

    def test_add0(self):
        assert self.dlt_filter.add("AAA", "BBB") == 0
        assert self.dlt_filter.counter == 1
        assert len(self.dlt_filter.apid[0]) == 4
        assert len(self.dlt_filter.ctid[0]) == 4
        assert ctypes.string_at(self.dlt_filter.apid[0], DLT_ID_SIZE) == b"AAA\0"
        assert ctypes.string_at(self.dlt_filter.ctid[0], DLT_ID_SIZE) == b"BBB\0"

    def test_add1(self):
        assert self.dlt_filter.add("AAA", "BBB") == 0
        assert self.dlt_filter.add("XXX", "YYY") == 0
        assert self.dlt_filter.counter == 2
        assert ctypes.string_at(self.dlt_filter.apid[0], DLT_ID_SIZE) == b"AAA\0"
        assert ctypes.string_at(self.dlt_filter.ctid[0], DLT_ID_SIZE) == b"BBB\0"
        assert ctypes.string_at(self.dlt_filter.apid[1], DLT_ID_SIZE) == b"XXX\0"
        assert ctypes.string_at(self.dlt_filter.ctid[1], DLT_ID_SIZE) == b"YYY\0"

    def test_add2(self):
        assert self.dlt_filter.add("AAAA", "BBBB") == 0
        assert self.dlt_filter.add("XXX", "YYY") == 0
        assert self.dlt_filter.add("CCCC", "DDDD") == 0
        assert self.dlt_filter.counter == 3
        assert ctypes.string_at(self.dlt_filter.apid[0], DLT_ID_SIZE) == b"AAAA"
        assert ctypes.string_at(self.dlt_filter.ctid[0], DLT_ID_SIZE) == b"BBBB"
        assert ctypes.string_at(self.dlt_filter.apid[1], DLT_ID_SIZE) == b"XXX\0"
        assert ctypes.string_at(self.dlt_filter.ctid[1], DLT_ID_SIZE) == b"YYY\0"
        assert ctypes.string_at(self.dlt_filter.apid[2], DLT_ID_SIZE) == b"CCCC"
        assert ctypes.string_at(self.dlt_filter.ctid[2], DLT_ID_SIZE) == b"DDDD"

    def test_repr(self):
        assert self.dlt_filter.add("AAAA", "BBBB") == 0
        assert self.dlt_filter.add("XXX", "YYY") == 0
        assert self.dlt_filter.add("CCCC", "DDDD") == 0
        print(self.dlt_filter)
        assert str(self.dlt_filter) == str([(b"AAAA", b"BBBB"), (b"XXX", b"YYY"), (b"CCCC", b"DDDD")])



# Copyright (C) 2021. BMW Car IT GmbH. All rights reserved.
"""Test DLTBroker with enabling dlt_time"""

from contextlib import contextmanager
from multiprocessing import Queue
import queue as tqueue
import time
from unittest.mock import ANY, patch, MagicMock

import pytest

from dlt.dlt_broker import create_filter_ack_queue, DLTBroker, logger
from dlt.dlt_broker_handlers import DLTContextHandler, DLTFilterAckMessageHandler, DLTMessageHandler
from tests.utils import MockDLTMessage


def fake_py_dlt_client_main_loop(client, callback, *args, **kwargs):
    return True


@contextmanager
def dlt_broker(pydlt_main_func=fake_py_dlt_client_main_loop, enable_dlt_time=True, enable_filter_set_ack=False):
    """Initialize a fake DLTBroker"""
    with (
        patch("dlt.dlt_broker_handlers.DLTMessageHandler._client_connect"),
        patch("dlt.dlt_broker_handlers.py_dlt_client_main_loop", side_effect=pydlt_main_func),
    ):
        broker = DLTBroker("42.42.42.42", enable_dlt_time=enable_dlt_time, enable_filter_set_ack=enable_filter_set_ack)
        broker.msg_handler._client = MagicMock()

        try:
            broker.start()

            yield broker

        finally:
            broker.stop()


@contextmanager
def dlt_filter_ack_msg_handler():
    queue = Queue()

    handler = DLTFilterAckMessageHandler(queue)
    try:
        handler.start()
        queue.cancel_join_thread()

        yield (handler, queue)
    finally:
        handler.stop()
        queue.close()


def fake_dlt_msg_handler(msg, with_filter_ack_queue):
    """Create a fake DLTMessageHandler"""
    filter_queue = MagicMock()
    filter_ack_queue = MagicMock() if with_filter_ack_queue else None
    client_cfg = {"ip_address": b"127.0.0.1", "filename": b"/dev/null", "verbose": 0, "port": "1234"}

    filter_queue.empty.side_effect = [False, True]
    filter_queue.get_nowait.return_value = msg

    return DLTMessageHandler(
        filter_queue, MagicMock(), MagicMock(), client_cfg, dlt_time_value=None, filter_ack_queue=filter_ack_queue
    )


def test_start_stop_dlt_broker():
    """Test to stop DLTBroker with dlt-time normally"""
    with dlt_broker(fake_py_dlt_client_main_loop, enable_dlt_time=True) as broker:
        assert broker._dlt_time_value


def test_start_stop_dlt_broker_without_dlt_time():
    """Test to stop DLTBroker without dlt-time normally"""
    with dlt_broker(fake_py_dlt_client_main_loop, enable_dlt_time=False) as broker:
        assert not broker._dlt_time_value


@pytest.mark.parametrize(
    "input_sec,input_msec,expected_value",
    [
        (42, 42, 42.42),  # normal test case
        (1618993559, 7377682, 1618993559.7377682),  # big value. The value will be truncated when type is not double
    ],
)
def test_dlt_broker_get_dlt_time(input_sec, input_msec, expected_value):
    """Test to get time from DLTBroker"""

    def handle(client, callback=None, *args, **kwargs):
        return callback(MockDLTMessage(payload="test_payload", sec=input_sec, msec=input_msec))

    with dlt_broker(handle) as broker:
        time.sleep(0.01)

    assert broker.dlt_time() == expected_value


def test_dlt_broker_get_latest_dlt_time():
    """Test to get the latest time from DLTBroker"""
    # ref: https://stackoverflow.com/questions/3190706/nonlocal-keyword-in-python-2-x
    time_value = {"v": 42}

    def handle(client, callback=None, *args, **kwargs):
        if time_value["v"] < 45:
            time_value["v"] += 1

        time.sleep(0.01)
        return callback(MockDLTMessage(payload="test_payload", sec=time_value["v"], msec=42))

    with dlt_broker(handle) as broker:
        time_vals = set()
        for i in range(10):
            time_vals.add(broker.dlt_time())
            time.sleep(0.01)

    assert sorted(time_vals) == [0.0, 43.42, 44.42, 45.42]


def test_start_stop_dlt_broker_with_dlt_ack_msg_handler():
    """Test to stop DLTBroker with ack msg handler normally"""
    with dlt_broker(fake_py_dlt_client_main_loop, enable_dlt_time=True, enable_filter_set_ack=True) as broker:
        assert broker.filter_ack_msg_handler


def test_start_stop_dlt_broker_without_dlt_ack_msg_handler():
    """Test to stop DLTBroker without ack msg handler normally"""
    with dlt_broker(fake_py_dlt_client_main_loop, enable_dlt_time=True, enable_filter_set_ack=False) as broker:
        assert not broker.filter_ack_msg_handler


def test_create_filter_ack_queue():
    """Test to register and unregister an ack queue"""
    handler_mock = MagicMock()

    with create_filter_ack_queue(handler_mock) as queue:
        queue.put(True)
        assert queue.get()

    handler_mock.register.assert_called_with(queue)
    handler_mock.unregister.assert_called_with(queue)


@pytest.mark.parametrize(
    "ack,required_ack,return_val",
    [
        (True, True, True),
        (False, False, True),
        (True, False, False),
        (False, True, False),
    ],
)
def test_recv_filter_set_ack(ack, required_ack, return_val):
    """Test to receive an ack value"""
    queue = tqueue.Queue()

    queue.put(ack)
    with dlt_broker(enable_filter_set_ack=True) as broker:
        assert return_val == broker._recv_filter_set_ack(queue, required_ack)


def test_recv_filter_set_ack_timeout_ignore():
    """Test not to receive an ack value"""
    queue = tqueue.Queue()

    with dlt_broker(enable_filter_set_ack=True) as broker:
        broker.filter_set_ack_timeout = 0.01
        broker.ignore_filter_set_ack_timeout = True

        assert not broker._recv_filter_set_ack(queue, True)


def test_recv_filter_set_ack_timeout_exception():
    """Test not to receive an ack value and with an exception"""
    queue = tqueue.Queue()

    with dlt_broker(enable_filter_set_ack=True) as broker:
        broker.filter_set_ack_timeout = 0.01
        broker.ignore_filter_set_ack_timeout = False

        with pytest.raises(tqueue.Empty) as err:
            broker._recv_filter_set_ack(queue, True)

        assert not str(err.value)


def test_add_context_with_ack():
    """Test to send a filter-setting message with required ack"""
    queue = tqueue.Queue()

    with patch("dlt.dlt_broker.DLTBroker._recv_filter_set_ack", return_value=True) as ack_mock:
        with dlt_broker(enable_filter_set_ack=True) as broker:
            ori_context_handler = broker.context_handler
            broker.context_handler = MagicMock()
            try:
                broker.add_context(queue, [("APID", "CTID")])

                broker.context_handler.register.assert_called()
                ack_mock.assert_called()
            finally:
                broker.context_handler = ori_context_handler


def test_add_context_with_ack_warning():
    """Test to send a filter-setting message but not received an ack"""
    queue = tqueue.Queue()

    with (
        patch("dlt.dlt_broker.DLTBroker._recv_filter_set_ack", return_value=False) as ack_mock,
        patch.object(logger, "warning") as logger_mock,
    ):
        with dlt_broker(enable_filter_set_ack=True) as broker:
            ori_context_handler = broker.context_handler
            broker.context_handler = MagicMock()
            try:
                broker.add_context(queue, [("APID", "CTID")])

                broker.context_handler.register.assert_called()
                ack_mock.assert_called()

                logger_mock.assert_called_with(ANY, ANY, [("APID", "CTID")], id(queue))
            finally:
                broker.context_handler = ori_context_handler


def test_start_stop_dlt_filter_ack_msg_handler():
    """Test to start/stop DLTFilterAckMessageHandler normally"""
    with dlt_filter_ack_msg_handler() as (handler, _):
        pass

    assert not handler.is_alive()


def test_dlt_filter_ack_msg_handler_register():
    """Test to register a new ack queue into DLTFilterAckMessageHandler"""
    queue_ack = tqueue.Queue()

    with dlt_filter_ack_msg_handler() as (handler, queue):
        handler.register(queue_ack)

        queue.put((id(queue_ack), True))
        assert queue_ack.get()


def test_dlt_filter_ack_msg_handler_unregister():
    """Test to unregister a new ack queue into DLTFilterAckMessageHandler"""
    queue_ack = tqueue.Queue()

    with dlt_filter_ack_msg_handler() as (handler, queue):
        handler.register(queue_ack)

        handler.unregister(queue_ack)
        with pytest.raises(tqueue.Empty):
            queue.put((id(queue_ack), False))
            queue_ack.get_nowait()


def test_make_send_filter_msg():
    """Test to generate a filter message"""
    handler = DLTContextHandler(MagicMock(), MagicMock())

    is_register = True
    filters = [("APID", "CTID")]
    queue = MagicMock()

    assert handler._make_send_filter_msg(queue, filters, is_register) == (id(queue), filters, is_register)


def test_make_send_filter_msg_with_ack_queue():
    """Test to generate a filter message with ack queue setting"""
    handler = DLTContextHandler(MagicMock(), MagicMock())

    is_register = True
    filters = [("APID", "CTID")]
    queue = MagicMock()
    queue_ack = MagicMock()

    assert handler._make_send_filter_msg(queue, filters, is_register, context_filter_ack_queue=queue_ack) == (
        id(queue),
        id(queue_ack),
        filters,
        is_register,
    )


def test_dlt_message_handler_process_filter_queue_add():
    """Test to add a filter"""
    handler = fake_dlt_msg_handler(msg=(42, [("APID", "CTID")], True), with_filter_ack_queue=True)
    handler._process_filter_queue()

    assert handler.context_map[("APID", "CTID")] == [42]
    handler.filter_ack_queue.put.assert_not_called()


def test_dlt_message_handler_process_filter_queue_add_ack():
    """Test to add a filter with ack"""
    handler = fake_dlt_msg_handler(msg=(42, 43, [("APID", "CTID")], True), with_filter_ack_queue=True)
    handler._process_filter_queue()

    assert handler.context_map[("APID", "CTID")] == [42]
    handler.filter_ack_queue.put.assert_called_with((43, True))


def test_dlt_message_handler_process_filter_queue_remove():
    """Test to remove a filter"""
    handler = fake_dlt_msg_handler(msg=(42, [("APID", "CTID")], False), with_filter_ack_queue=True)
    handler.context_map[("APID", "CTID")].append(42)

    handler._process_filter_queue()

    assert ("APID", "CTID") not in handler.context_map
    handler.filter_ack_queue.put.assert_not_called()


def test_dlt_message_handler_process_filter_queue_remove_ack():
    """Test to remove a filter with ack"""
    handler = fake_dlt_msg_handler(msg=(42, 43, [("APID", "CTID")], False), with_filter_ack_queue=True)
    handler.context_map[("APID", "CTID")].append(42)

    handler._process_filter_queue()

    assert ("APID", "CTID") not in handler.context_map
    handler.filter_ack_queue.put.assert_called_with((43, False))


def test_dlt_message_handler_process_filter_queue_remove_exception():
    """Test to remove a filter when the filter does not exists"""
    handler = fake_dlt_msg_handler(msg=(42, [("APID", "CTID")], False), with_filter_ack_queue=True)

    handler._process_filter_queue()

    assert not handler.context_map[("APID", "CTID")]
    handler.filter_ack_queue.put.assert_not_called()



# Copyright (C) 2017. BMW Car IT GmbH. All rights reserved.
"""Basic ctypes binding to the DLT library"""

import ctypes
import os

from dlt.core.core_base import *  # noqa: F403

API_VER = None


def get_version(loaded_lib):
    """Return the API version of the loaded libdlt.so library"""
    global API_VER  # pylint: disable=global-statement
    if API_VER is None:
        buf = ctypes.create_string_buffer(255)
        loaded_lib.dlt_get_version(ctypes.byref(buf), 255)
        # buf would be something like:
        # DLT Package Version: X.XX.X STABLE, Package Revision: vX.XX.XX build on Jul XX XXXX XX:XX:XX
        # -SYSTEMD -SYSTEMD_WATCHDOG -TEST -SHM
        buf_split = buf.value.decode().split()

        API_VER = buf_split[3]

    return API_VER


def get_api_specific_file(version):
    """Return specific version api filename, if not found fallback to highest matching version"""
    version_tuple = [int(num) for num in version.split(".")]
    name = "core_{}.py".format("".join((str(num) for num in version_tuple)))
    base_dir = os.path.dirname(os.path.abspath(__file__))
    if os.path.exists(os.path.join(base_dir, name)):
        return name

    # Fallback logic: Try to find the closest core_*.py
    for i in range(len(version_tuple)):
        truncated_tuple = version_tuple[: -(i + 1)] + [0] * (i + 1)
        name = "core_{}.py".format("".join((str(num) for num in truncated_tuple)))
        if os.path.exists(os.path.join(base_dir, name)):
            return name

    # If still not found, try to find the highest major version
    name = "core_{}00.py".format(version_tuple[0])
    if os.path.exists(os.path.join(base_dir, name)):
        return name

    raise ImportError("No module file: {}".format(name))


def check_libdlt_version(api_ver):
    """Check the version compatibility.

    python-dlt now only supports to run libdlt 2.18.5 or above.
    """
    ver_info = tuple(int(num) for num in api_ver.split("."))
    if ver_info < (2, 18, 5):
        raise ImportError("python-dlt only supports libdlt \
        v2.18.5 (33fbad18c814e13bd7ba2053525d8959fee437d1) or above")


API_VER = get_version(dltlib)  # noqa: F405
check_libdlt_version(API_VER)

# Load version specific definitions, if such a file exists, possibly
# overriding above definitions
#
# The intent is to have version specific implementations to be able to
# provide declarations *incrementally*.
#
# For instance if version 2.17.0 introduces new changes in addition to
# retaining all changes from 2.16.0, then core_2170.py would import
# core_2160.py and declare only version specific changes/overrides. The
# loading logic here below should not require changes.
#
# This allows the implementation below to import just one final module
# (as opposed to loading multiple implementations in a specific order)
# to provide new/overriding implementations.
api_specific_file = get_api_specific_file(API_VER)
overrides = __import__("dlt.core.{}".format(api_specific_file[:-3]), globals(), locals(), ["*"])
locals().update(overrides.__dict__)



# Copyright (C) 2015. BMW Car IT GmbH. All rights reserved.
"""DLT support module"""

import collections
import logging
import subprocess

if not hasattr(subprocess, "TimeoutExpired"):
    import subprocess32 as subprocess  # pylint: disable=import-error


LOGGER = logging.getLogger(__name__)
ProcessResult = collections.namedtuple("ProcessResult", ("stdout", "stderr", "returncode"))


def run_command(command, timeout=60, shell=True):
    """Run command in a shell and return stdout, stderr and return code

    :param str|list command: a command to run
    :param int timeout: timeout for the command
    :param bool shell: shell switch
    :returns: process result
    :rtype: subprocess compatible ProcessResult
    :raises RuntimeError: If timeout expires.
    """
    process = subprocess.Popen(
        command, shell=shell, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    try:
        stdout, stderr = process.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        process.terminate()
        raise RuntimeError("Timeout %d seconds reached for command '%s'" % (timeout, command))
    if isinstance(stdout, bytes):
        stdout = stdout.decode("utf-8")
    if isinstance(stderr, bytes):
        stderr = stderr.decode("utf-8")
    return ProcessResult(stdout, stderr, process.returncode)



# Copyright (C) 2015. BMW Car IT GmbH. All rights reserved.
"""DLT Broker is running in a loop in a separate thread until stop_flag is set and adding received messages
to all registered queues
"""

from contextlib import contextmanager
import ipaddress as ip
import logging
from multiprocessing import Event, Queue
import queue as tqueue

from dlt.dlt_broker_handlers import (
    DLT_DAEMON_TCP_PORT,
    DLTContextHandler,
    DLTFilterAckMessageHandler,
    DLTMessageDispatcherBase,
    DLTMessageHandler,
    DLTFileSpinner,
    DLTTimeValue,
)

DLT_CLIENT_TIMEOUT = 5

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@contextmanager
def create_filter_ack_queue(filter_ack_msg_handler):
    """Register and unregister a queue into DLTFilterAckMessageHandler"""
    queue = tqueue.Queue()

    try:
        filter_ack_msg_handler.register(queue)

        yield queue

    finally:
        filter_ack_msg_handler.unregister(queue)


class DLTBroker(object):
    """DLT Broker class manages receiving and filtering of DLT Messages"""

    def __init__(
        self,
        ip_address=None,
        port=DLT_DAEMON_TCP_PORT,
        use_proxy=False,
        enable_dlt_time=False,
        enable_filter_set_ack=False,
        filter_set_ack_timeout=2.0,
        ignore_filter_set_ack_timeout=False,
        **kwargs,
    ):
        """Initialize the DLT Broker

        :param str | None ip_address: IP address of the DLT Daemon.
                If None, then dlt does not come with any ip listening, in other words, it comes from dlt log directly;
                Else, dlt comes from listening to some ip address. Defaults to TCP connection,
                unless a multicast address is used. In that case an UDP multicast connection will be used
        :param str post: Port of the DLT Daemon
        :param bool use_proxy: Ignored - compatibility option
        :param bool enable_dlt_time: Record the latest dlt message timestamp if enabled.
        :param bool enable_filter_set_ack: Wait an ack message when sending a filter-setting message
        :param float filter_set_ack_timeout: Waiting time for the ack message
        :param bool ignore_filter_set_ack_timeout: Ignore the timeout when the value is True
        :param **kwargs: All other args passed to DLTMessageHandler
        """
        # - dlt-time share memory init
        self._dlt_time_value = DLTTimeValue() if enable_dlt_time else None

        # - handlers init
        self.mp_stop_flag = Event()
        self.filter_queue = Queue()
        self.message_queue = Queue()

        # - filter ack queue setting
        self.enable_filter_set_ack = enable_filter_set_ack
        self.ignore_filter_set_ack_timeout = ignore_filter_set_ack_timeout
        self.filter_set_ack_timeout = filter_set_ack_timeout
        if enable_filter_set_ack:
            # Optional[multiprocessing.Queue[Tuple[int, bool]]]
            # int presents queue id, bool presents enable or not
            self.filter_ack_queue = Queue()
            self.filter_ack_msg_handler = DLTFilterAckMessageHandler(self.filter_ack_queue)
        else:
            self.filter_ack_queue = None
            self.filter_ack_msg_handler = None

        self.msg_handler = self.create_dlt_message_dispather(ip_address, port, kwargs)

        self.context_handler = DLTContextHandler(self.filter_queue, self.message_queue)

        self._ip_address = ip_address
        self._port = port
        self._filename = kwargs.get("filename")

    def create_dlt_message_dispather(self, ip_address, port, client_cfg) -> DLTMessageDispatcherBase:
        if ip_address:
            # If ip_address is given, then messages are retrieved from dlt client at run-time
            return self._create_dlt_message_handler(ip_address, port, client_cfg)
        else:
            # If not ip_address is given, then messages are retrieved from the given filename
            # The logs are written to the given filename from another process
            return self._create_dlt_file_spinner(client_cfg.get("filename"))

    def _create_dlt_message_handler(self, ip_address, port, client_cfg):
        client_cfg["ip_address"] = ip_address
        client_cfg["port"] = port
        client_cfg["timeout"] = client_cfg.get("timeout", DLT_CLIENT_TIMEOUT)
        return DLTMessageHandler(
            self.filter_queue,
            self.message_queue,
            self.mp_stop_flag,
            client_cfg,
            dlt_time_value=self._dlt_time_value,
            filter_ack_queue=self.filter_ack_queue,
        )

    def _create_dlt_file_spinner(self, file_name):
        return DLTFileSpinner(
            self.filter_queue,
            self.message_queue,
            self.mp_stop_flag,
            file_name,
            dlt_time_value=self._dlt_time_value,
            filter_ack_queue=self.filter_ack_queue,
        )

    def start(self):
        """DLTBroker main worker method"""
        if isinstance(self.msg_handler, DLTMessageHandler):
            logger.debug(
                "Starting DLTBroker with parameters: use_proxy=%s, ip_address=%s, port=%s, filename=%s, multicast=%s",
                False,
                self._ip_address,
                self._port,
                self._filename,
                ip.ip_address(self._ip_address).is_multicast,
            )
        else:
            logger.debug("Starting DLTBroker by reading %s", self._filename)

        if self._dlt_time_value:
            logger.debug("Enable dlt time for DLTBroker.")

        self.msg_handler.start()
        self.context_handler.start()
        if self.enable_filter_set_ack:
            self.filter_ack_msg_handler.start()

        # - ensure we don't block on join_thread() in stop()
        # https://docs.python.org/2.7/library/multiprocessing.html#multiprocessing.Queue.cancel_join_thread
        self.filter_queue.cancel_join_thread()
        self.message_queue.cancel_join_thread()
        if self.enable_filter_set_ack:
            self.filter_ack_queue.cancel_join_thread()

    def _recv_filter_set_ack(self, context_filter_ack_queue, required_response):
        try:
            resp = context_filter_ack_queue.get(timeout=self.filter_set_ack_timeout)
            if resp != required_response:
                logger.debug("Filter-setting ack response not matched: %s, expected: %s", resp, required_response)
                return False

            return True
        except tqueue.Empty as err:
            if self.ignore_filter_set_ack_timeout:
                logger.info(
                    "Timeout for getting filter-setting ack: %s, %s", id(context_filter_ack_queue), required_response
                )
                return None

            raise err

        return False

    def add_context(self, context_queue, filters=None):
        """Register context

        :param Queue context_queue: The queue to which new messages will
                                    be added
        :param tuple filters: An list of tuples (eg: [(apid, ctid)])
                              used to filter messages that go into this
                              queue.
        """
        filters = filters or [(None, None)]

        if not isinstance(filters, (tuple, list)):
            raise RuntimeError("Context queue filters must be a tuple. Ex. (('SYS', 'JOUR'), ('AUDI', 'CAPI'))")

        if self.enable_filter_set_ack:
            logger.debug("Send a filter-setting message with requesting ack")
            with create_filter_ack_queue(self.filter_ack_msg_handler) as context_filter_ack_queue:
                self.context_handler.register(
                    context_queue, filters, context_filter_ack_queue=context_filter_ack_queue
                )

                if not self._recv_filter_set_ack(context_filter_ack_queue, True):
                    failure_reason = ""
                    if isinstance(self.msg_handler, DLTMessageHandler):
                        failure_reason = (
                            "It's possible that DLTClient client does not start."
                            " If it's a test case, it might be an error"
                        )
                    elif isinstance(self.msg_handler, DLTFileSpinner):
                        failure_reason = (
                            f"It's possible that dlt file {self._filename} is empty now. No big issue, "
                            f"filters would be added once after new message is available in dlt file"
                        )
                    logger.warning(
                        (
                            "Could not receive filter-setting message ack. %s. For now, Run it anyway. "
                            "filters: %s, queue_id: %s"
                        ),
                        failure_reason,
                        filters,
                        id(context_queue),
                    )
        else:
            self.context_handler.register(context_queue, filters)

    def remove_context(self, context_queue):
        """Unregister context

        :param Queue context_queue: The queue to unregister.
        """
        self.context_handler.unregister(context_queue)

    def stop(self):
        """Stop the broker"""
        logger.info("Stopping DLTContextHandler and %s", type(self.msg_handler).__name__)

        self.msg_handler.break_blocking_main_loop()

        logger.debug("Stop %s", type(self.msg_handler).__name__)
        self.mp_stop_flag.set()

        logger.debug("Stop DLTContextHandler")
        self.context_handler.stop()

        logger.debug("Waiting on DLTContextHandler ending")
        self.context_handler.join()

        if self.enable_filter_set_ack:
            logger.debug("Stop DLTFilterAckMessageHandler")
            self.filter_ack_msg_handler.stop()

            logger.debug("Waiting on DLTFilterAckMessageHandler ending")
            self.filter_ack_msg_handler.join()

        logger.debug("Waiting on %s ending", type(self.msg_handler).__name__)
        if self.msg_handler.is_alive():
            try:
                self.msg_handler.terminate()
            except OSError:
                pass
            else:
                self.msg_handler.join()

        logger.debug("DLTBroker execution done")

    # pylint: disable=invalid-name
    def isAlive(self):
        """Backwards compatibility method

        Called from mtee.testing.connectors.tools.broker_assert. Will
        need to be replaced in MTEE eventually.
        """
        return any((self.msg_handler.is_alive(), self.context_handler.is_alive()))

    def dlt_time(self):
        """Get time for the last dlt message

        The value is seconds from 1970/1/1 0:00:00

        :rtype: float
        """
        if self._dlt_time_value:
            return self._dlt_time_value.timestamp

        raise RuntimeError("Getting dlt time function is not enabled")

    @property
    def injection_support(self) -> bool:
        return bool(self._ip_address)



# Copyright (C) 2015. BMW Car IT GmbH. All rights reserved.
"""Handlers are classes that assist dlt_broker in receiving and
filtering DLT messages
"""

from abc import ABC, abstractmethod
from collections import defaultdict
import ctypes
import logging
from multiprocessing import Lock, Process, Value
from queue import Empty
import socket
import time
from threading import Thread, Event

from dlt.dlt import (
    DLTClient,
    DLT_DAEMON_TCP_PORT,
    DLT_UDP_MULTICAST_BUFFER_SIZE,
    DLT_UDP_MULTICAST_FD_BUFFER_SIZE,
    cDLT_FILE_NOT_OPEN_ERROR,
    load,
    py_dlt_client_main_loop,
    py_dlt_file_main_loop,
)

DLT_CLIENT_TIMEOUT = 5
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class DLTTimeValue(object):
    """Create a share memory to pass the timestamp between processes

    The type of dlt time is float (4 bytes). There are several ways to send
    the value between DLTMessageHandler (it's a process) and DLTBroker. Since
    DLTMessageHandler has to send the value many times every second, choosing a
    lightweight solution is must.

    khiz678 studied and evaluated the following solutions for the problem.
      - multiprocessing.Queue (Queue in the following discussion)
      - multiprocessing.Pipe (Pipe in the following discussion)
      - multiprocessing.Value (Value in the following discussion)

    Value is our final solution. Queue's implementation is based on Pipe (in
    cpython). If the solution is based on Queue or Pipe, it needs another
    thread in DLTBroker process to receive and write the value to a local
    variable. The solution based on Value does not have such problem, only
    assigns the value to the shared memory directly.

    khiz678 also did a simple benchmark for the Value solution. It could
    receive more than 100000 timestamps per second.  It's twice faster than
    Pipe's implementation.
    """

    def __init__(self, default_value=0.0):
        self._timestamp_mem = Value(ctypes.c_double, default_value)

    @property
    def timestamp(self):
        """Get the seconds from 1970/1/1 0:00:00

        :rtype: float
        """
        with self._timestamp_mem.get_lock():
            return self._timestamp_mem.value

    @timestamp.setter
    def timestamp(self, new_timestamp):
        with self._timestamp_mem.get_lock():
            self._timestamp_mem.value = new_timestamp


class DLTFilterAckMessageHandler(Thread):
    """Receive filter-set ack message and pass it to the corresponding ack queue"""

    def __init__(self, filter_ack_queue):  # (multiprocessing.Queue[Tuple[ContextQueueID, bool]]) -> None
        super(DLTFilterAckMessageHandler, self).__init__()

        self.filter_ack_queue = filter_ack_queue

        self.stop_flag = Event()

        self.context_map_lock = Lock()
        self.context_map = {}  # Dict[ContextQueueID, Queue.Queue[bool]]

    def stop(self):
        """Stops thread execution"""
        self.stop_flag.set()
        self.filter_ack_queue.put((None, None))
        if self.is_alive():
            self.join()

    def register(self, filter_ack_queue):  # (Queue.Queue[bool]) -> None
        """Register an ack queue"""
        with self.context_map_lock:
            self.context_map[id(filter_ack_queue)] = filter_ack_queue

    def unregister(self, filter_ack_queue):  # (Queue.Queue[bool]) -> None
        """Unregister an ack queue"""
        with self.context_map_lock:
            key = id(filter_ack_queue)

            if key in self.context_map:
                del self.context_map[key]

    def run(self):
        """Run the thread to recieve the message and dispatch it"""
        while not self.stop_flag.is_set():
            queue_ack_id, enable = self.filter_ack_queue.get()

            if not queue_ack_id:
                continue

            with self.context_map_lock:
                if queue_ack_id in self.context_map:
                    self.context_map[queue_ack_id].put(enable)
                else:
                    logger.warning("Could not send an ack to the queue: %s", queue_ack_id)


class DLTContextHandler(Thread):
    """Communication layer between the DLTContext instances and
    DLTMessageHandler Process.

    This class handles the transfer of messages between the process
    receiving traces from the DLT Daemon and the DLTContext queues.
    """

    def __init__(self, filter_queue, message_queue):
        super(DLTContextHandler, self).__init__()
        self.stop_flag = Event()
        self.context_map = {}
        self.lock = Lock()
        self.filter_queue = filter_queue
        self.message_queue = message_queue

    def _make_send_filter_msg(self, queue, filters, is_register, context_filter_ack_queue=None):
        """Send a filter message to the filter message queue"""
        queue_id = id(queue)

        if context_filter_ack_queue:
            return queue_id, id(context_filter_ack_queue), filters, is_register

        return queue_id, filters, is_register

    def register(self, queue, filters=None, context_filter_ack_queue=None):
        """Register a queue to collect messages matching specific filters

        :param Queue queue: The new queue to add
        :param tuple filters: An tuple with (apid, ctid) used to filter
                              messages that go into this queue.
        """
        if filters is None:
            filters = [(None, None)]

        queue_id = id(queue)  # - unique identifier for this queue
        with self.lock:
            self.context_map[queue_id] = (queue, filters)

        # - inform the DLTMessageHandler process about this new
        # (queue, filter) pair
        self.filter_queue.put(
            self._make_send_filter_msg(queue, filters, True, context_filter_ack_queue=context_filter_ack_queue)
        )

    def unregister(self, queue, context_filter_ack_queue=None):
        """Remove a queue from set of queues being handled

        :param Queue queue: The queue to remove
        """
        queue_id = id(queue)
        _, filters = self.context_map.get(queue_id, (None, None))
        if filters:
            with self.lock:
                try:
                    del self.context_map[queue_id]
                except KeyError:
                    pass

            # - inform the DLTMessageHandler process about removal of this
            # (queue, filter) pair
            self.filter_queue.put(
                self._make_send_filter_msg(queue, filters, False, context_filter_ack_queue=context_filter_ack_queue)
            )

    def run(self):
        """The thread's main loop"""
        while not self.stop_flag.is_set():
            queue_id, message = None, None
            try:
                if self.message_queue.full():
                    logger.error("message_queue is full ! put() on this queue will block")
                queue_id, message = self.message_queue.get_nowait()
            except Empty:
                pass

            if message:
                queue, _ = self.context_map.get(queue_id, (None, None))
                if queue:
                    queue.put(message)
            else:
                time.sleep(0.01)

    def stop(self):
        """Stops thread execution"""
        self.stop_flag.set()
        self.filter_queue.close()
        if self.is_alive():
            self.join()


class DLTMessageDispatcherBase(ABC, Process):
    """Base class for different dlt message dispatchers

    The derived class could dispatch dlt messages from dlt-daemon, or from at-runtime written file.
    """

    def __init__(self, filter_queue, message_queue, mp_stop_event, dlt_time_value=None, filter_ack_queue=None):
        """
        Common members needed for common dispatching behavirours

        :param Queue filter_queue: contexts for filtering received dlt message
        :param Queue message_queue: received dlt messages after filtering against context
        :param multiprocessing.Event mp_stop_event: stop signal for this process
        :param bool enable_dlt_time: Record the latest dlt message timestamp if enabled.
        :param bool filter_ack_queue: acks for accepting contexts
        """
        super().__init__()
        self.filter_queue = filter_queue
        self.filter_ack_queue = filter_ack_queue
        self.message_queue = message_queue
        self.mp_stop_flag = mp_stop_event
        # - dict mapping filters to queue ids
        self.context_map = defaultdict(list)
        self._dlt_time_value = dlt_time_value

    def _process_filter_queue(self):
        """Check if filters have been added or need to be removed"""
        while not self.filter_queue.empty():
            queue_ack_id = None

            msg = self.filter_queue.get_nowait()
            logger.debug("Process filter queue message: %s", msg)
            if isinstance(msg, tuple) and len(msg) == 4:
                queue_id, queue_ack_id, filters, add = msg
            else:
                queue_id, filters, add = msg

            if add:
                for apid_ctid in filters:
                    self.context_map[apid_ctid].append(queue_id)
            else:
                try:
                    for apid_ctid in filters:
                        self.context_map[apid_ctid].remove(queue_id)
                        if not self.context_map[apid_ctid]:
                            del self.context_map[apid_ctid]
                except (KeyError, ValueError):
                    # - queue_id already removed or not inserted
                    pass

            if self.filter_ack_queue and queue_ack_id:
                logger.debug("Send filter ack message: queue_ack_id: %s, add: %s", queue_ack_id, add)
                self.filter_ack_queue.put((queue_ack_id, add))

    @abstractmethod
    def is_valid_message(self, message):
        """Validate if the received message is a valid message according to AUTOSAR doc"""
        return True

    def handle(self, message):
        """Function to be called for every message received

        :param DLTMessage message: received new DLTMessage instance
        :returns: True if the loop should continue, False to stop the loop and exit
        :rtype: bool
        """
        self._process_filter_queue()

        if self.is_valid_message(message):
            # Dispatch the message
            msg_ctx = ((message.apid, message.ctid), (None, None), (message.apid, None), (None, message.ctid))
            qids = (
                queue_id
                for filters, queue_ids in self.context_map.items()
                for queue_id in queue_ids
                if filters in msg_ctx
            )
            for queue_id in qids:
                if self.message_queue.full():
                    logger.error("message_queue is full ! put() on this queue will block")
                self.message_queue.put((queue_id, message))

            # Send the message's timestamp
            if self._dlt_time_value:
                self._dlt_time_value.timestamp = message.storage_timestamp

        return not self.mp_stop_flag.is_set()

    @abstractmethod
    def run(self) -> None:
        pass

    def break_blocking_main_loop(self):
        """All message dispatchers need a main loop to fetch dlt messages from source.
        If it could constantly dispatch messages, then the main loop will not get into blocking state.
        Only when no more message could not be dispatched, the main loop would get into blocking state.

        Not all message dispatchers need to implement this method
        """
        pass


class DLTFileSpinner(DLTMessageDispatcherBase):
    """Process receiving the DLT messages and handing them to DLTContextHandler

    This process instance is responsible for collecting messages from
    the at-runtime written dlt log, tagging them with the correct queue id and placing
    them on the messages queue.
    """

    def __init__(
        self, filter_queue, message_queue, mp_stop_event, file_name, dlt_time_value=None, filter_ack_queue=None
    ):
        super().__init__(filter_queue, message_queue, mp_stop_event, dlt_time_value, filter_ack_queue)
        self.file_name = file_name
        self.dlt_reader = load(filename=self.file_name, live_run=True)

    def is_valid_message(self, message):
        """According to AUTOSAR doc, message with empty apid and empty ctid is still valid"""
        return message is not None

    def run(self):
        """DLTFileSpinner worker method"""
        logger.info("Start to process dlt file %s", self.file_name)
        # Even though dlt connector for ioc should only be instantiated after successful SerialConsole with fibex,
        # the corner case of not-existing dlt file will still be handled here with max 5 retires
        retries_for_non_existing_file = 5

        while not self.mp_stop_flag.is_set():
            try:
                logger.debug("py_dlt_file_main_loop")
                res = py_dlt_file_main_loop(self.dlt_reader, callback=self.handle)
                if res is False and not self.mp_stop_flag.is_set():  # main loop returned False
                    logger.error("Too many bad messages read from %s", self.file_name)
                    self.mp_stop_flag.set()
                    break
            except KeyboardInterrupt:
                logger.debug("main loop manually interrupted")
                break
            except IOError as err:
                if str(err) == cDLT_FILE_NOT_OPEN_ERROR:
                    # Not every time of non-existing file, cDLTFile will report error
                    # Sometimes, it simply works through without issue.
                    # So, no unittest could be done for this error handling
                    if retries_for_non_existing_file == 0:
                        logger.error("After retries, dlt file %s still does not exist", self.file_name)
                        raise err
                    logger.warning(
                        "DLT file %s does not exist, will try %d times again",
                        self.file_name,
                        retries_for_non_existing_file,
                    )
                    retries_for_non_existing_file = retries_for_non_existing_file - 1
                    time.sleep(1)
                else:
                    raise err
            except Exception:  # pylint: disable=broad-except
                logger.exception("Exception during the DLT message receive")

        logger.debug("DLTFileSpinner starts to quit...")
        if not self.dlt_reader.stop_reading_proc.is_set():
            self.dlt_reader.stop_reading_proc.set()
        self.message_queue.close()
        logger.info("DLTFileSpinner worker execution complete")

    def break_blocking_main_loop(self):
        """A big user for DLTFileSpinner is IOC dlt, which does not have so many dlt messages as HU,
        so it is quite easy for the main loop to get into blocking state,
        at the moment that no more dlt messages could be dispatched.
        """
        logger.debug("Stop iterating to file %s", self.file_name)
        self.dlt_reader.stop_reading_proc.set()


class DLTMessageHandler(DLTMessageDispatcherBase):
    """Process receiving the DLT messages and handing them to DLTContextHandler

    This process instance is responsible for collecting messages from
    the DLT daemon, tagging them with the correct queue id and placing
    them on the messages queue.
    """

    def __init__(
        self, filter_queue, message_queue, mp_stop_event, client_cfg, dlt_time_value=None, filter_ack_queue=None
    ):
        super().__init__(filter_queue, message_queue, mp_stop_event, dlt_time_value, filter_ack_queue)
        self._ip_address = client_cfg["ip_address"]
        self._port = client_cfg.get("port", DLT_DAEMON_TCP_PORT)
        self._filename = client_cfg.get("filename")
        self.verbose = client_cfg.get("verbose", 0)
        self.timeout = client_cfg.get("timeout", DLT_CLIENT_TIMEOUT)
        self._client = None
        self.tracefile = None
        self.last_connected = time.time()
        self.last_message = time.time() - 120.0
        self._udp_fd_buffer_size_bytes = client_cfg.get("udp_fd_buffer_size_bytes", DLT_UDP_MULTICAST_FD_BUFFER_SIZE)
        self._udp_buffer_size_bytes = client_cfg.get("udp_buffer_size_bytes", DLT_UDP_MULTICAST_BUFFER_SIZE)

    def is_valid_message(self, message):
        return message and (message.apid != "" or message.ctid != "")

    def _client_connect(self):
        """Create a new DLTClient

        :param int timeout: Time in seconds to wait for connection.
        :returns: True if connected, False otherwise
        :rtype: bool
        """
        if self.verbose:
            logger.debug(
                "Creating DLTClient (ip_address='%s', Port='%s', logfile='%s')",
                self._ip_address,
                self._port,
                self._filename,
            )
        self._client = DLTClient(
            servIP=self._ip_address,
            port=self._port,
            verbose=self.verbose,
            udp_fd_buffer_size_bytes=self._udp_fd_buffer_size_bytes,
            udp_buffer_size_bytes=self._udp_buffer_size_bytes,
        )
        connected = self._client.connect(self.timeout)
        if connected:
            logger.info("DLTClient connected to %s", self._client.servIP)
        return connected

    def run(self, reconnect_delay=0.1):
        """DLTMessageHandler worker method"""
        if self._filename is not None:
            logger.info("Opening the DLT trace file '%s'", self._filename)
            self.tracefile = open(self._filename, mode="ab", buffering=False)

        is_reconnect = False
        while not self.mp_stop_flag.is_set():
            exception_occured = False
            if not self._client_connect():
                # keep trying to reconnect, until we either successfully
                # connect or the stop_flag is set
                if time.time() - self.last_message > 60:
                    # Once per minute log that we still have no DLT Connection
                    logger.info(
                        "DLT connection to %s missing since %s seconds",
                        self._ip_address,
                        time.time() - self.last_connected,
                    )
                    self.last_message = time.time()
                time.sleep(reconnect_delay)
                continue
            try:
                if self.last_connected:
                    logger.info(
                        "DLT connection to %s re-established after %s seconds",
                        self._ip_address,
                        time.time() - self.last_connected,
                    )
                if is_reconnect:
                    time.sleep(reconnect_delay)
                    is_reconnect = False
                self.last_connected = time.time()
                res = py_dlt_client_main_loop(self._client, verbose=0, callback=self.handle, dumpfile=self.tracefile)
                if res is False and not self.mp_stop_flag.is_set():  # main loop returned False
                    logger.warning("DLT connection to %s lost. Restarting DLT client", self._ip_address)
                    self.last_connected = time.time()
                    self.last_message = time.time()
                    exception_occured = True
                    is_reconnect = True
            except KeyboardInterrupt:
                exception_occured = True
                logger.debug("main loop manually interrupted")
                break
            except socket.timeout as exc:
                exception_occured = True
                logger.error("socket timeout error")
                logger.debug(exc)
            except Exception:  # pylint: disable=broad-except
                exception_occured = True
                logger.exception("Exception during the DLT message receive")

            finally:
                if exception_occured:
                    logger.debug("Closing open socket connections.")
                    self._client.disconnect()

        self.message_queue.close()
        self._client.disconnect()
        logger.info("DLTMessageHandler worker execution complete")



# Copyright (C) 2016. BMW Car IT GmbH. All rights reserved.
from multiprocessing import Queue as mp_queue
from queue import Empty, Queue
import time
import unittest

from dlt.dlt_broker_handlers import DLTContextHandler
from tests.utils import create_messages, stream_one, stream_multiple


class TestDLTContextHandler(unittest.TestCase):
    def setUp(self):
        self.filter_queue = mp_queue()
        self.message_queue = mp_queue()
        self.handler = DLTContextHandler(self.filter_queue, self.message_queue)

    def test_init(self):
        self.assertFalse(self.handler.stop_flag.is_set())
        self.assertFalse(self.handler.is_alive())
        self.assertTrue(self.handler.filter_queue.empty())
        self.assertTrue(self.handler.message_queue.empty())

    def test_register_no_filter(self):
        queue = Queue()
        queue_id = id(queue)

        self.handler.register(queue)

        # When no filter is specified, filter (None, None) should be
        # added (ie: match all messages)
        self.assertIn(queue_id, self.handler.context_map)
        self.assertEqual(self.handler.context_map[queue_id], (queue, [(None, None)]))
        self.assertEqual(self.handler.filter_queue.get(), (queue_id, [(None, None)], True))

    def test_register_single_filter(self):
        queue = Queue()
        queue_id = id(queue)
        filters = ("SYS", "JOUR")

        self.handler.register(queue, filters)

        # Specified, filter should be added to filter_queue
        self.assertIn(queue_id, self.handler.context_map)
        self.assertEqual(self.handler.context_map[queue_id], (queue, filters))
        self.assertEqual(self.handler.filter_queue.get(), (queue_id, filters, True))

    def test_register_similar_filters(self):
        queue0 = Queue()
        queue_id0 = id(queue0)
        filters0 = ("SYS", "JOUR")

        queue1 = Queue()
        queue_id1 = id(queue1)
        filters1 = ("SYS", "JOUR")

        self.handler.register(queue0, filters0)
        self.handler.register(queue1, filters1)

        # Each queue should have a unique entry in the context_map and
        # filter_queue (even if they have the same filter)
        self.assertIn(queue_id0, self.handler.context_map)
        self.assertIn(queue_id1, self.handler.context_map)
        self.assertEqual(self.handler.context_map[queue_id0], (queue0, filters0))
        self.assertEqual(self.handler.context_map[queue_id1], (queue1, filters1))
        self.assertEqual(self.handler.filter_queue.get(), (queue_id0, filters0, True))
        self.assertEqual(self.handler.filter_queue.get(), (queue_id1, filters1, True))

    def test_unregister(self):
        queue = Queue()
        queue_id = id(queue)
        filters = ("SYS", "JOUR")

        self.handler.register(queue, filters)
        self.assertIn(queue_id, self.handler.context_map)
        self.assertEqual(self.handler.filter_queue.get(), (queue_id, filters, True))

        self.handler.unregister(queue)
        self.assertNotIn(queue_id, self.handler.context_map)
        self.assertEqual(self.handler.filter_queue.get(), (queue_id, filters, False))

    def test_run_no_messages(self):
        try:
            self.handler.start()
            time.sleep(0.2)
            self.handler.stop()
            self.assertTrue(self.handler.stop_flag.is_set())
            self.assertFalse(self.handler.is_alive())
        except:  # noqa: E722
            self.fail()

    def test_run_single_context_queue(self):
        queue = Queue()
        queue_id = id(queue)
        filters = ("DA1", "DC1")
        self.handler.register(queue, filters)

        self.handler.start()

        # - simulate feeding of messages into the message_queue
        for _ in range(10):
            self.handler.message_queue.put((queue_id, create_messages(stream_one)))

        try:
            for _ in range(10):
                queue.get(timeout=0.01)
        except Empty:
            # - we should not get an Empty for exactly 10 messages
            self.fail()
        finally:
            self.handler.stop()

    def test_run_multiple_context_queue(self):
        self.handler.start()

        queue0 = Queue()
        queue_id0 = id(queue0)
        filters0 = ("DA1", "DC1")
        self.handler.register(queue0, filters0)

        queue1 = Queue()
        queue_id1 = id(queue1)
        filters1 = ("SYS", "JOUR")
        self.handler.register(queue1, filters1)

        # - queue with no filter
        queue2 = Queue()
        queue_id2 = id(queue2)
        self.handler.register(queue2)

        # - simulate feeding of messages into the message_queue
        for _ in range(10):
            for message in create_messages(stream_multiple, from_file=True):
                queue_id = queue_id0 if message.apid == "DA1" else queue_id1
                self.handler.message_queue.put((queue_id, message))
                # - simulate feeding of all messages for the queue with
                # no filter.
                self.handler.message_queue.put((queue_id2, message))

        try:
            da1_messages = []
            sys_messages = []
            all_messages = []
            for _ in range(10):
                da1_messages.append(queue0.get(timeout=0.01))
                sys_messages.append(queue1.get(timeout=0.01))
                all_messages.append(queue2.get(timeout=0.01))

            # these queues should not get any messages from other queues
            self.assertTrue(all(msg.apid == "DA1" for msg in da1_messages))
            self.assertTrue(all(msg.apid == "SYS" for msg in sys_messages))
            # this queues should get all messages
            self.assertFalse(
                all(msg.apid == "DA1" for msg in all_messages) or all(msg.apid == "SYS" for msg in all_messages)
            )
        except Empty:
            # - we should not get an Empty for at least 10 messages
            self.fail()
        finally:
            self.handler.stop()

    def test_run_unregister_with_unread_messages(self):
        self.handler.start()
        queue = Queue()
        queue_id = id(queue)
        filters = ("DA1", "DC1")
        self.handler.register(queue, filters)

        self.assertIn(queue_id, self.handler.context_map)
        self.handler.unregister(queue)

        # - simulate feeding of messages into the message_queue
        for _ in range(3):
            self.handler.message_queue.put((queue_id, create_messages(stream_one)))

        try:
            self.assertNotIn(queue_id, self.handler.context_map)
            # allow some time for the thread to read all messages
            time.sleep(0.5)
            self.assertTrue(self.handler.message_queue.empty())
            self.assertTrue(queue.empty())
        finally:
            self.handler.stop()



# Copyright (C) 2023. BMW Car IT GmbH. All rights reserved.
"""Test DLTBroker with message handler DLTFileSpinner"""

import os
import pytest
import tempfile
import time
import unittest
from unittest.mock import ANY, patch
from queue import Queue, Empty

from dlt.dlt_broker import DLTBroker, logger
from tests.utils import (
    stream_multiple,
    stream_with_params,
    append_stream_to_file,
    create_messages,
    append_message_to_file,
)


class TestDLTBrokerFromDLTFileSpinnerWithNotExistingDLT(unittest.TestCase):
    def setUp(self) -> None:
        self.broker = None
        _, self.dlt_file_name = tempfile.mkstemp(suffix=b".dlt")

    def tearDown(self) -> None:
        if self.broker:
            self.broker.stop()
        if os.path.exists(self.dlt_file_name):
            os.remove(self.dlt_file_name)

    def test_broker_with_not_existing_dlt_file(self):
        """
        Test DLTBroker could work with not existing dlt file

        1. prepare a file name which does not exist
        2. start dlt broker to dispatch messages from this not-existing file --> no error
        3. dlt broker could not add context successfully, but encounter a warning message
        4. no message could be dispatched from not existing file and throws out Queue.Empty exception
        5. dlt_time is 0.0, because it could not be reset according to the latest timestamp of messages
        """
        # Remove the dlt file
        os.remove(self.dlt_file_name)
        # Start broker with non-existing dlt file
        self.broker = DLTBroker(
            filename=self.dlt_file_name,
            enable_dlt_time=True,
            enable_filter_set_ack=True,
            ignore_filter_set_ack_timeout=True,
        )
        self.broker.start()
        # Add context should report warning message
        queue = Queue(maxsize=0)
        with patch.object(logger, "warning") as logger_mock:
            self.broker.add_context(queue, filters=None)
            logger_mock.assert_called_with(ANY, ANY, [(None, None)], id(queue))
        # Not existing dlt file should not throw any exception out
        for _ in range(5):
            with pytest.raises(Empty):
                queue.get_nowait()
        # dlt_time is not None, even though it is not reset with latest timestamp from messages
        self.assertEqual(self.broker.dlt_time(), 0.0)

    def test_broker_with_later_created_dlt_file(self):
        """
        Simulate a scenario: first dlt file does not exist, then dlt file is created and written with messages.

        1. delete the dlt file
        2. start broker
        3. create the dlt file and write 1 sample message
            Expectation: 1 message could be dispatched from broker
        """
        # 1. delete the dlt file
        os.remove(self.dlt_file_name)
        # 2. Start broker with non-existing dlt file
        self.broker = DLTBroker(
            filename=self.dlt_file_name,
            enable_dlt_time=True,
            enable_filter_set_ack=True,
            ignore_filter_set_ack_timeout=True,
        )
        self.broker.start()
        # Add context should report warning message
        queue = Queue(maxsize=0)
        self.broker.add_context(queue, filters=None)
        # 3. Write 1 sample message to the dlt file
        append_stream_to_file(stream_with_params, self.dlt_file_name)
        # Expectation: 1 message could be dispatched from broker
        time.sleep(0.5)
        self.assertIsNotNone(queue.get_nowait())
        # If we try to dispatch for another time, exception Queue.Empty is thrown,
        # because there is no new log from dlt file
        with pytest.raises(Empty):
            queue.get_nowait()


class TestDLTBrokerFromDLTFileSpinner(unittest.TestCase):
    def setUp(self):
        # Dlt file is created with empty content
        _, self.dlt_file_name = tempfile.mkstemp(suffix=b".dlt")
        self.dispatched_message_queue = Queue(maxsize=0)
        # Instantiate DLTBroker without ignoring fileter ack timeout
        self.broker = DLTBroker(
            filename=self.dlt_file_name,
            enable_dlt_time=True,
            enable_filter_set_ack=True,
            ignore_filter_set_ack_timeout=True,
        )
        self.broker.start()
        self.broker.add_context(self.dispatched_message_queue, filters=None)

    def tearDown(self):
        self.broker.stop()
        os.remove(self.dlt_file_name)

    def test_001_dispatch_from_empty_dlt_file(self):
        """
        From empty file, no message could be dispatched from queue and raise Queue.Empty.
        dlt_time is 0.0, because it could not be reset according to the latest timestamp of messages
        """
        for _ in range(5):
            with pytest.raises(Empty):
                self.dispatched_message_queue.get_nowait()
        self.assertEqual(self.broker.dlt_time(), 0.0)

    def test_002_dispatch_from_real_dlt_file(self):
        """
        Test DltBroker dispatches from a run-time written dlt file

        With a running dlt broker:
        1. Write 2 sample messages to dlt file
        2. These two messages could be dispatched with the running dlt broker
           With another try to dispatch, Queue.Empty is thrown, because no more logs could be read from dlt log;
           dlt_time from dlt_broker is equal to the timestamp of 2nd message
        3. Append another 1 message to the same dlt file
        4. Total 3 messages could be dispatched with the dlt broker
           With another try to dispatch, Queue.Empty is thrown, because no more logs could be read from dlt log;
           dlt_time from dlt_broker is equal to the timestamp of 3rd message
        """
        # 1. Write 2 sample messages to dlt file
        append_stream_to_file(stream_multiple, self.dlt_file_name)
        # 2. Dispatch 2 messages from dlt broker
        time.sleep(0.1)
        message_1 = self.dispatched_message_queue.get_nowait()
        time.sleep(0.1)
        message_2 = self.dispatched_message_queue.get_nowait()
        self.assertNotEqual(message_1, message_2)
        # If we try to dispatch for another time, exception Queue.Empty is thrown,
        # because there is no new log from dlt file
        with pytest.raises(Empty):
            self.dispatched_message_queue.get_nowait()
        # Validate dlt time from broker
        self.assertEqual(self.broker.dlt_time(), message_2.storage_timestamp)
        # 3. Append another 1 message to the same dlt file
        append_stream_to_file(stream_with_params, self.dlt_file_name)
        # 4. Total 3 messages could be dispatched with the dlt broker
        time.sleep(0.1)
        message_3 = self.dispatched_message_queue.get_nowait()
        self.assertNotEqual(message_1, message_3)
        self.assertNotEqual(message_2, message_3)
        # If try to dispatch for another time, exception Queue.Empty is thrown,
        # because there is no new log from dlt file
        with pytest.raises(Empty):
            self.dispatched_message_queue.get_nowait()
        # Validate dlt time from broker
        self.assertEqual(self.broker.dlt_time(), message_3.storage_timestamp)

    def test_003_dispatch_from_real_dlt_file(self):
        """
        Test DltBroker dispatches apid==b"" and ctid==b"" message from a run-time written dlt file

        With a running dlt broker:
        1. Write apid==b"" and ctid==b"" message to dlt file
        2. This message could be dispatched with the running dlt broker
           a. With another try to dispatch, Queue.Empty is thrown, because no more logs could be read from dlt log;
           b. dlt_time from dlt_broker is equal to the timestamp of this message
           c. the received message should have apid==b"" and ctid==b""
        """
        # 1. Write apid==b"" and ctid==b"" message to dlt file
        # Construct a message with apid==b"" and ctid==b""
        message = create_messages(stream_with_params, from_file=True)[0]
        message.extendedheader.apid = b""
        message.extendedheader.ctid = b""
        # Write this message into dlt file
        append_message_to_file(message, self.dlt_file_name)
        # 2. Dispatch from dlt broker
        time.sleep(0.5)
        message = self.dispatched_message_queue.get_nowait()
        # If we try to dispatch for another time, exception Queue.Empty is thrown,
        # because there is no new log from dlt file
        with pytest.raises(Empty):
            self.dispatched_message_queue.get_nowait()
        # Validate dlt time from broker
        self.assertEqual(self.broker.dlt_time(), message.storage_timestamp)
        # Expectation: the received message should have apid==b"" and ctid==b""
        self.assertEqual("", message.apid)
        self.assertEqual("", message.ctid)



# Copyright (C) 2016. BMW Car IT GmbH. All rights reserved.
"""Basic unittests for the py_dlt_client_main_loop function"""

import ctypes
import functools
from io import BytesIO as StringIO
import socket
import unittest
from unittest.mock import patch, Mock

from dlt.dlt import py_dlt_client_main_loop, DLTClient, logger
from dlt.core import cDltStorageHeader
from tests.utils import stream_one


def mock_dlt_receiver_receive_socket(client_receiver, partial=False, Fail=False):
    if Fail:
        return 0
    stream_one.seek(0)
    buf = stream_one.read()
    if partial:
        buf = buf[:16]

    client_receiver._obj.buf = ctypes.create_string_buffer(buf)
    client_receiver._obj.bytesRcvd = len(buf)
    return len(buf)


class TestMainLoopWithDltClient(unittest.TestCase):
    def setUp(self):
        self.client = DLTClient()
        self.client._connected_socket = Mock()

    def test_target_down(self):
        with patch.object(self.client._connected_socket, "recv", side_effect=socket.timeout):
            callback = Mock(return_value="should not be called")

            with self.assertLogs(logger=logger) as dlt_logger:
                return_value = py_dlt_client_main_loop(self.client, callback=callback)
                self.assertFalse(return_value)

                log_output = dlt_logger.output
                self.assertEqual(len(log_output), 1)
                self.assertEqual(log_output[0], "ERROR:dlt.dlt:[]: DLTLib closed connected socket")

            self.assertFalse(callback.called)

    def test_target_up_nothing_to_read(self):
        with patch.object(self.client._connected_socket, "recv", return_value=b"") as mock_recv:
            callback = Mock(return_value="should not be called")
            self.assertFalse(py_dlt_client_main_loop(self.client, callback=callback))
            self.assertEqual(mock_recv.call_count, 1)
            self.assertFalse(callback.called)

    @patch("dlt.dlt.dltlib.dlt_receiver_move_to_begin", return_value=0)
    def test_exit_if_callback_returns_false(self, *ignored):
        with patch.object(self.client._connected_socket, "recv", return_value=b"X"):
            # setup dlt_receiver_receive to return a partial message
            replacement = functools.partial(mock_dlt_receiver_receive_socket, partial=True)
            with patch("dlt.dlt.dltlib.dlt_receiver_receive", new=replacement):
                self.assertFalse(py_dlt_client_main_loop(self.client, callback=lambda msg: False))

    def test_read_message(self, *ignored):
        dumpfile = StringIO()

        stream_one.seek(0)
        expected = stream_one.read()

        with patch.object(self.client._connected_socket, "recv", return_value=b"X"):
            # setup dlt_receiver_receive to return a complete message
            replacement = functools.partial(mock_dlt_receiver_receive_socket)
            callback = Mock(side_effect=[True, False, False])
            with patch("dlt.dlt.dltlib.dlt_receiver_receive", new=replacement):
                self.assertTrue(py_dlt_client_main_loop(self.client, dumpfile=dumpfile, callback=callback))
                self.assertEqual(dumpfile.getvalue()[ctypes.sizeof(cDltStorageHeader) :], expected)
